@@ -2,6 +2,7 @@ package release
 
 import (
 	"bytes"
+	"context"
 	"encoding/json"
 	"fmt"
 	"io"
@@ -21,8 +22,8 @@ import (
 	"k8s.io/cli-runtime/pkg/genericclioptions"
 	"k8s.io/client-go/util/retry"
 	"k8s.io/klog"
-	kcmdutil "k8s.io/kubernetes/pkg/kubectl/cmd/util"
-	"k8s.io/kubernetes/pkg/kubectl/util/templates"
+	kcmdutil "k8s.io/kubectl/pkg/cmd/util"
+	"k8s.io/kubectl/pkg/util/templates"
 	"sigs.k8s.io/yaml"
 
 	imagev1 "github.com/openshift/api/image/v1"
@@ -31,6 +32,7 @@ import (
 	"github.com/openshift/library-go/pkg/image/dockerv1client"
 	imagereference "github.com/openshift/library-go/pkg/image/reference"
 	"github.com/openshift/oc/pkg/cli/image/extract"
+	"github.com/openshift/oc/pkg/cli/image/imagesource"
 	imagemanifest "github.com/openshift/oc/pkg/cli/image/manifest"
 	"github.com/openshift/oc/pkg/cli/image/mirror"
 )
@@ -70,7 +72,24 @@ func NewMirror(f kcmdutil.Factory, parentName string, streams genericclioptions.
 			correct information to give to OpenShift to use that content offline. An alternate mode
 			is to specify --to-image-stream, which imports the images directly into an OpenShift
 			image stream.
+
+			You may use --to-dir to specify a directory to download release content into, and add
+			the file:// prefix to the --to flag. The command will print the 'oc image mirror' command
+			that can be used to upload the release to another registry.
 		`),
+		Example: templates.Examples(`
+			# Perform a dry run showing what would be mirrored, including the mirror objects
+			%[1]s 4.2.2 --to myregistry.local/openshift/release --dry-run
+
+			# Mirror a release into the current directory
+			%[1]s 4.2.2 --to file://openshift/release
+
+			# Mirror a release to another directory in the default location
+			%[1]s 4.2.2 --to-dir /tmp/releases
+
+			# Upload a release from the current directory to another server
+			%[1]s --from file://openshift/release --to myregistry.com/openshift/release
+			`),
 		Run: func(cmd *cobra.Command, args []string) {
 			kcmdutil.CheckErr(o.Complete(cmd, f, args))
 			kcmdutil.CheckErr(o.Run())
@@ -83,6 +102,9 @@ func NewMirror(f kcmdutil.Factory, parentName string, streams genericclioptions.
 	flags.StringVar(&o.From, "from", o.From, "Image containing the release payload.")
 	flags.StringVar(&o.To, "to", o.To, "An image repository to push to.")
 	flags.StringVar(&o.ToImageStream, "to-image-stream", o.ToImageStream, "An image stream to tag images into.")
+	flags.StringVar(&o.FromDir, "from-dir", o.ToDir, "A directory to import images from.")
+	flags.StringVar(&o.ToDir, "to-dir", o.ToDir, "A directory to export images to.")
+	flags.BoolVar(&o.ToMirror, "to-mirror", o.ToMirror, "Output the mirror mappings instead of mirroring.")
 	flags.BoolVar(&o.DryRun, "dry-run", o.DryRun, "Display information about the mirror without actually executing it.")
 
 	flags.BoolVar(&o.SkipRelease, "skip-release-image", o.SkipRelease, "Do not push the release image.")
@@ -96,14 +118,19 @@ type MirrorOptions struct {
 	SecurityOptions imagemanifest.SecurityOptions
 	ParallelOptions imagemanifest.ParallelOptions
 
-	From string
+	From    string
+	FromDir string
 
 	To            string
 	ToImageStream string
 
+	// modifies the targets
 	ToRelease   string
 	SkipRelease bool
 
+	ToMirror bool
+	ToDir    string
+
 	DryRun                        bool
 	PrintImageContentInstructions bool
 
@@ -124,6 +151,16 @@ func (o *MirrorOptions) Complete(cmd *cobra.Command, f kcmdutil.Factory, args []
 	case len(args) > 1:
 		return fmt.Errorf("only one argument is accepted")
 	}
+
+	args, err := findArgumentsFromCluster(f, []string{o.From})
+	if err != nil {
+		return err
+	}
+	if len(args) != 1 {
+		return fmt.Errorf("only one release image may be mirrored")
+	}
+	o.From = args[0]
+
 	o.ClientFn = func() (imageclient.Interface, string, error) {
 		cfg, err := f.ToRESTConfig()
 		if err != nil {
@@ -144,13 +181,31 @@ func (o *MirrorOptions) Complete(cmd *cobra.Command, f kcmdutil.Factory, args []
 }
 
 const replaceComponentMarker = "X-X-X-X-X-X-X"
+const replaceVersionMarker = "V-V-V-V-V-V-V"
 
 func (o *MirrorOptions) Run() error {
 	if len(o.From) == 0 && o.ImageStream == nil {
 		return fmt.Errorf("must specify a release image with --from")
 	}
 
-	if (len(o.To) == 0) == (len(o.ToImageStream) == 0) {
+	outputs := 0
+	if len(o.To) > 0 {
+		outputs++
+	}
+	if len(o.ToImageStream) > 0 {
+		outputs++
+	}
+	if len(o.ToDir) > 0 {
+		if outputs == 0 {
+			outputs++
+		}
+	}
+	if o.ToMirror {
+		if outputs == 0 {
+			outputs++
+		}
+	}
+	if outputs != 1 {
 		return fmt.Errorf("must specify an image repository or image stream to mirror the release to")
 	}
 
@@ -160,7 +215,7 @@ func (o *MirrorOptions) Run() error {
 
 	var recreateRequired bool
 	var hasPrefix bool
-	var targetFn func(name string) mirror.MirrorReference
+	var targetFn func(name string) imagesource.TypedImageReference
 	var dst string
 	if len(o.ToImageStream) > 0 {
 		dst = imagereference.DockerImageReference{
@@ -171,35 +226,55 @@ func (o *MirrorOptions) Run() error {
 	} else {
 		dst = o.To
 	}
+	if len(dst) == 0 {
+		if len(o.ToDir) > 0 {
+			dst = "file://openshift/release"
+		} else {
+			dst = "openshift/release"
+		}
+	}
 
+	var toDisk bool
+	var version string
 	if strings.Contains(dst, "${component}") {
 		format := strings.Replace(dst, "${component}", replaceComponentMarker, -1)
-		dstRef, err := mirror.ParseMirrorReference(format)
+		format = strings.Replace(format, "${version}", replaceVersionMarker, -1)
+		dstRef, err := imagesource.ParseReference(format)
 		if err != nil {
 			return fmt.Errorf("--to must be a valid image reference: %v", err)
 		}
-		targetFn = func(name string) mirror.MirrorReference {
+		toDisk = dstRef.Type == imagesource.DestinationFile
+		targetFn = func(name string) imagesource.TypedImageReference {
+			if len(name) == 0 {
+				name = "release"
+			}
 			value := strings.Replace(dst, "${component}", name, -1)
-			ref, err := mirror.ParseMirrorReference(value)
+			value = strings.Replace(value, "${version}", version, -1)
+			ref, err := imagesource.ParseReference(value)
 			if err != nil {
 				klog.Fatalf("requested component %q could not be injected into %s: %v", name, dst, err)
 			}
 			return ref
 		}
 		replaceCount := strings.Count(dst, "${component}")
-		recreateRequired = replaceCount > 1 || (replaceCount == 1 && !strings.Contains(dstRef.Tag, replaceComponentMarker))
+		recreateRequired = replaceCount > 1 || (replaceCount == 1 && !strings.Contains(dstRef.Ref.Tag, replaceComponentMarker))
 
 	} else {
-		ref, err := mirror.ParseMirrorReference(dst)
+		ref, err := imagesource.ParseReference(dst)
 		if err != nil {
 			return fmt.Errorf("--to must be a valid image repository: %v", err)
 		}
-		if len(ref.ID) > 0 || len(ref.Tag) > 0 {
+		toDisk = ref.Type == imagesource.DestinationFile
+		if len(ref.Ref.ID) > 0 || len(ref.Ref.Tag) > 0 {
 			return fmt.Errorf("--to must be to an image repository and may not contain a tag or digest")
 		}
-		targetFn = func(name string) mirror.MirrorReference {
+		targetFn = func(name string) imagesource.TypedImageReference {
 			copied := ref
-			copied.Tag = name
+			if len(name) > 0 {
+				copied.Ref.Tag = fmt.Sprintf("%s-%s", version, name)
+			} else {
+				copied.Ref.Tag = version
+			}
 			return copied
 		}
 		hasPrefix = true
@@ -207,7 +282,7 @@ func (o *MirrorOptions) Run() error {
 
 	o.TargetFn = func(name string) imagereference.DockerImageReference {
 		ref := targetFn(name)
-		return ref.DockerImageReference
+		return ref.Ref
 	}
 
 	if recreateRequired {
@@ -245,31 +320,67 @@ func (o *MirrorOptions) Run() error {
 			fmt.Fprintf(o.ErrOut, "warning: %v\n", err)
 		}
 	}
+	version = is.Name
 
+	// sourceFn is given a chance to rewrite source mappings
+	sourceFn := func(ref imagesource.TypedImageReference) imagesource.TypedImageReference {
+		return ref
+	}
 	var mappings []mirror.Mapping
-	if len(o.From) > 0 && !o.SkipRelease {
+	if len(o.From) > 0 {
 		src := o.From
-		srcRef, err := imagereference.Parse(src)
+		srcRef, err := imagesource.ParseReference(src)
 		if err != nil {
-			return err
+			return fmt.Errorf("invalid --from: %v", err)
 		}
+
+		// if the source ref is a file type, provide a function that checks the local file store for a given manifest
+		// before continuing, to allow mirroring an entire release to disk in a single file://REPO.
+		if srcRef.Type == imagesource.DestinationFile {
+			if repo, err := (&imagesource.Options{FileDir: o.FromDir}).Repository(context.TODO(), srcRef); err == nil {
+				sourceFn = func(ref imagesource.TypedImageReference) imagesource.TypedImageReference {
+					if ref.Type == imagesource.DestinationFile || len(ref.Ref.ID) == 0 {
+						return ref
+					}
+					manifests, err := repo.Manifests(context.TODO())
+					if err != nil {
+						klog.V(2).Infof("Unable to get local manifest service: %v", err)
+						return ref
+					}
+					ok, err := manifests.Exists(context.TODO(), digest.Digest(ref.Ref.ID))
+					if err != nil {
+						klog.V(2).Infof("Unable to get check for local manifest: %v", err)
+						return ref
+					}
+					if !ok {
+						return ref
+					}
+					updated := srcRef
+					updated.Ref.Tag = ""
+					updated.Ref.ID = ref.Ref.ID
+					klog.V(2).Infof("Rewrote %s to %s", ref, updated)
+					return updated
+				}
+			} else {
+				klog.V(2).Infof("Unable to build local file lookup: %v", err)
+			}
+		}
+
 		if len(o.ToRelease) > 0 {
-			dstRef, err := imagereference.Parse(o.ToRelease)
+			dstRef, err := imagesource.ParseReference(o.ToRelease)
 			if err != nil {
 				return fmt.Errorf("invalid --to-release-image: %v", err)
 			}
 			mappings = append(mappings, mirror.Mapping{
-				Type:        mirror.DestinationRegistry,
 				Source:      srcRef,
 				Destination: dstRef,
 				Name:        o.ToRelease,
 			})
 		} else if !o.SkipRelease {
-			dstRef := targetFn("release")
+			dstRef := targetFn("")
 			mappings = append(mappings, mirror.Mapping{
 				Source:      srcRef,
-				Type:        dstRef.Type(),
-				Destination: dstRef.Combined(),
+				Destination: dstRef,
 				Name:        "release",
 			})
 		}
@@ -291,29 +402,39 @@ func (o *MirrorOptions) Run() error {
 			return fmt.Errorf("image-references should only contain pointers to images by digest: %s", tag.From.Name)
 		}
 
+		// Allow mirror refs to be sourced locally
+		srcMirrorRef := imagesource.TypedImageReference{Ref: from, Type: imagesource.DestinationRegistry}
+		srcMirrorRef = sourceFn(srcMirrorRef)
+
 		// Create a unique map of repos as keys
 		currentRepo := from.AsRepository().String()
 		repositories[currentRepo] = struct{}{}
 
 		dstMirrorRef := targetFn(tag.Name)
 		mappings = append(mappings, mirror.Mapping{
-			Source:      from,
-			Type:        dstMirrorRef.Type(),
-			Destination: dstMirrorRef.Combined(),
+			Source:      srcMirrorRef,
+			Destination: dstMirrorRef,
 			Name:        tag.Name,
 		})
 		klog.V(2).Infof("Mapping %#v", mappings[len(mappings)-1])
 
 		dstRef := targetFn(tag.Name)
-		dstRef.Tag = ""
-		dstRef.ID = from.ID
-		tag.From.Name = dstRef.Exact()
+		dstRef.Ref.Tag = ""
+		dstRef.Ref.ID = from.ID
+		tag.From.Name = dstRef.Ref.Exact()
 	}
 
 	if len(mappings) == 0 {
 		fmt.Fprintf(o.ErrOut, "warning: Release image contains no image references - is this a valid release?\n")
 	}
 
+	if o.ToMirror {
+		for _, mapping := range mappings {
+			fmt.Fprintf(o.Out, "%s %s\n", mapping.Source.String(), mapping.Destination.String())
+		}
+		return nil
+	}
+
 	if len(o.ToImageStream) > 0 {
 		remaining := make(map[string]mirror.Mapping)
 		for _, mapping := range mappings {
@@ -346,10 +467,13 @@ func (o *MirrorOptions) Run() error {
 					},
 				}
 				for _, mapping := range remaining {
+					if mapping.Source.Type != imagesource.DestinationRegistry {
+						return fmt.Errorf("source mapping %s must point to a registry", mapping.Source)
+					}
 					isi.Spec.Images = append(isi.Spec.Images, imagev1.ImageImportSpec{
 						From: corev1.ObjectReference{
 							Kind: "DockerImage",
-							Name: mapping.Source.Exact(),
+							Name: mapping.Source.Ref.Exact(),
 						},
 						To: &corev1.LocalObjectReference{
 							Name: mapping.Name,
@@ -415,6 +539,7 @@ func (o *MirrorOptions) Run() error {
 	opts.SecurityOptions = o.SecurityOptions
 	opts.ParallelOptions = o.ParallelOptions
 	opts.Mappings = mappings
+	opts.FileDir = o.ToDir
 	opts.DryRun = o.DryRun
 	opts.ManifestUpdateCallback = func(registry string, manifests map[digest.Digest]digest.Digest) error {
 		lock.Lock()
@@ -447,23 +572,33 @@ func (o *MirrorOptions) Run() error {
 
 	to := o.ToRelease
 	if len(to) == 0 {
-		to = targetFn("release").String()
+		to = targetFn("").Ref.Exact()
 	}
-	if hasPrefix {
-		fmt.Fprintf(o.Out, "\nSuccess\nUpdate image:  %s\nMirror prefix: %s\n", to, o.To)
-	} else {
-		fmt.Fprintf(o.Out, "\nSuccess\nUpdate image:  %s\nMirrored to: %s\n", to, o.To)
+	fmt.Fprintf(o.Out, "\nSuccess\nUpdate image:  %s\n", to)
+	if len(o.To) > 0 {
+		if hasPrefix {
+			fmt.Fprintf(o.Out, "Mirror prefix: %s\n", o.To)
+		} else {
+			fmt.Fprintf(o.Out, "Mirrored to: %s\n", o.To)
+		}
 	}
-
-	if o.PrintImageContentInstructions {
-		if err := printImageContentInstructions(o.Out, o.From, o.To, repositories); err != nil {
-			return fmt.Errorf("Error creating mirror usage instructions: %v", err)
+	if toDisk {
+		if len(o.ToDir) > 0 {
+			fmt.Fprintf(o.Out, "\nTo upload local images to a registry, run:\n\n    oc image mirror --from-dir=%s 'file://%s*' REGISTRY/REPOSITORY\n\n", o.ToDir, to)
+		} else {
+			fmt.Fprintf(o.Out, "\nTo upload local images to a registry, run:\n\n    oc image mirror 'file://%s*' REGISTRY/REPOSITORY\n\n", to)
+		}
+	} else if len(o.To) > 0 {
+		if o.PrintImageContentInstructions {
+			if err := printImageContentInstructions(o.Out, o.From, o.To, repositories); err != nil {
+				return fmt.Errorf("Error creating mirror usage instructions: %v", err)
+			}
 		}
 	}
 	return nil
 }
 
-// printImageContentInstructions provides exapmles to the user for using the new repository mirror
+// printImageContentInstructions provides examples to the user for using the new repository mirror
 // https://github.com/openshift/installer/blob/master/docs/dev/alternative_release_image_sources.md
 func printImageContentInstructions(out io.Writer, from, to string, repositories map[string]struct{}) error {
 	type installConfigSubsection struct {
@@ -472,18 +607,24 @@ func printImageContentInstructions(out io.Writer, from, to string, repositories
 
 	var sources []operatorv1alpha1.RepositoryDigestMirrors
 
-	mirrorRef, err := imagereference.Parse(to)
+	mirrorRef, err := imagesource.ParseReference(to)
 	if err != nil {
 		return fmt.Errorf("Unable to parse image reference '%s': %v", to, err)
 	}
-	mirrorRepo := mirrorRef.AsRepository().String()
+	if mirrorRef.Type != imagesource.DestinationRegistry {
+		return nil
+	}
+	mirrorRepo := mirrorRef.Ref.AsRepository().String()
 
 	if len(from) != 0 {
-		sourceRef, err := imagereference.Parse(from)
+		sourceRef, err := imagesource.ParseReference(from)
 		if err != nil {
 			return fmt.Errorf("Unable to parse image reference '%s': %v", from, err)
 		}
-		sourceRepo := sourceRef.AsRepository().String()
+		if sourceRef.Type != imagesource.DestinationRegistry {
+			return nil
+		}
+		sourceRepo := sourceRef.Ref.AsRepository().String()
 		repositories[sourceRepo] = struct{}{}
 	}
 
